import os
import sys
import time
from pprint import pprint

import numpy as np
import pandas as pd
# import quandl
import scipy.optimize as sco
from funpicker import Query, QueryTypes
from funtime import Store
from pypfopt import expected_returns, risk_models
from pypfopt.efficient_frontier import EfficientFrontier

MONGO = os.getenv('MONGODB', "localhost")
PORTFOLIO_COLLECTION = os.getenv("PORTFOLIO_COLLECTION", "global")

# from portfolio_server.config import PORTFOLIO_COLLECTION, LOCALMONGO
#Converter

store = Store(MONGO).create_lib(PORTFOLIO_COLLECTION).get_store()
priceStore = store[PORTFOLIO_COLLECTION]


def get_portfolio(portfolio_list):
    general_timeperiod = (60 * 60) * 24 # 25 hours
    coin_dict = {

    }
    for coin in portfolio_list:
        latest = priceStore.query_latest({'type': 'price', 'exchange': "CCCAGG", "period" : "day", "trade": coin, "base" : "USD", "limit": 25})
        latest = list(latest)
        # print(latest)
        if len(latest) != 0:
            # print("it's not zero")
            # Check to see if the price was pulled in the last 18 hours
            most_recent_price = latest[0]

            current_time = time.time()
            time_check = general_timeperiod + most_recent_price['timestamp']
            # print(time_check, current_time)
            if time_check > current_time:
                # Add coin to dict
                # print("The time is totally right")
                coin_dict[coin] = latest
                # print(coin_dict)
                continue
            
        print(coin, " -- Getting Checked")
        fpq = Query().set_crypto(coin).set_fiat("USD").set_exchange("CCCAGG").set_period("day").set_limit(25).get()
        coin_dict[coin] = []
        for item in fpq:
            item['timestamp'] = float(item.pop('time'))
            item['type'] = "price"
            item['period'] = "day"
            item['exchange'] = "CCCAGG"
            item['trade'] = coin
            item['base'] = "USD"
            priceStore.store(item)
            coin_dict[coin].append(item)
        
    return coin_dict


def get_coin_pricing(exchange, base, trade):
    general_timeperiod = (1800) # 1 minute
    coin_dict = {

    }

    latest = priceStore.query({'type': 'price', 'exchange': exchange, "period" : "minute", "trade": trade, "base" : base})
    latest = list(latest)
    if len(latest) > 0:
        time_check = general_timeperiod + latest[0]['timestamp']
        current_time = time.time()
        if time_check > current_time:
            return latest[0]
    fpq = Query().set_crypto(trade).set_fiat(base).set_exchange(exchange).set_period("minute").set_limit(5).get()
    
    if len(fpq) == 0:
        return {
            "type": "price",
            "period": "minute",
            "timestamp": time.time(),
            "trade": trade,
            "base": base,
            "exchange": exchange,
            "close": 0
        }
    
    for item in fpq:
        item['timestamp'] = float(item.pop('time'))
        item['type'] = "price"
        item['period'] = "minute"
        item['exchange'] = exchange
        item['trade'] = trade
        item['base'] = base
        priceStore.store(item)
    
    return fpq[0]

def create_portfolio_price_df(coin_dict):
    portfolio_frame = pd.DataFrame()
    for k, v in coin_dict.items():
        if k is None:
            print("Don't do anything")
            continue
        df = pd.DataFrame(v)
        ts = pd.to_datetime(df['timestamp'], unit='s')
        df.reindex(pd.DatetimeIndex(ts))
        portfolio_frame[k] = df['close'] 
    return portfolio_frame.sort_index()




def portfolio_from_df(df, risk_lvl=0):
    mu = expected_returns.mean_historical_return(df)
    S = risk_models.sample_cov(df)

    # # Optimise for maximal Sharpe ratio
    ef = EfficientFrontier(mu, S, gamma=risk_lvl)
    raw_weights = ef.max_sharpe()
    cleaned_weights = ef.clean_weights()
    ef.portfolio_performance(verbose=True)

    return cleaned_weights



import maya
import copy
from maya import MayaDT
from abc import ABC
from link_jam_handle.data.base_handlers import DataSet

class RunnableSet(DataSet, ABC):
    """ Create a setup to run through code as if it were real (backtest/simulations). Might migrate that code to the portfolio. """
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._parameters = {}
        self._start_time_maya = maya.now()
        self._current_time = None
    

    @property
    def current_time(self):
        """ Get the timestamp of the price indicator. """
        step = self.current_step
        shift = 1
        
        return (self._start_time_maya.add(
            seconds=(self._seconds * (step+self._first_limit-shift)), 
            minutes=(self._minutes * (step+self._first_limit-shift)), 
            hours=(self._hours * (step+self._first_limit-shift)), 
            days=(self._days * (step+self._first_limit-shift)), 
            months=(self._months * (step+self._first_limit-shift)), 
            years=(self._years * (step+self._first_limit-shift))
        )._epoch)
        """ 
            return (self._start_time_maya.add(
                seconds=(self._seconds * (step+self._first_limit-shift)), 
                minutes=(self._minutes * (step+self._first_limit-shift)), 
                hours=(self._hours * (step+self._first_limit-shift)), 
                days=(self._days * (step+self._first_limit-shift)), 
                months=(self._months * (step+self._first_limit-shift)), 
                years=(self._years * (step+self._first_limit-shift))
            )._epoch)
        """

    @property
    def start_time(self):
        return self._start_time_maya._epoch
    

    @property
    def current_step(self):
        return self._step_count()

    @property
    def parameters(self):
        """ Parameters to we use to generate the data. """
        return self._parameters


    def update_step(self):
        """ Update the step of the program """
        self._save_step()


    @start_time.setter
    def start_time(self, _t:MayaDT):
        self.start_time = _t
    



    def _save_step(self):
        alt = {"details": "step"}
        data = {"count": "count"}
        self.save(data, alt=alt)


    def _step_count(self):
        alt = {"details": "step"}
        count = self.count(alt=alt)
        return count


    @parameters.setter
    def parameters(self, param:dict):
        """ Add items to the parameter dict. """
        keys = param.keys()
        for key in keys:
            self._parameters[key] = param[key]
    
    

    def step(self):
        """ Step through simulated data """
        raise NotImplementedError

    def reset(self):
        """ Run reset command to signify that we're loading information and"""
        raise NotImplementedError

    def close(self):
        raise NotImplementedError


from abc import ABC
from link_jam_handle.data.base_handlers import DataSet
class PriceSet(DataSet, ABC):
    """ A price dataset abstract. Use with all price related sets """
    def __init__(self, **kwargs) -> None:
        print(kwargs)
        super().__init__(**kwargs)
    

    def latest_price(self, symbol:str):
        """ Get the latest price """
        return 0.0
    
    def step(self):
        return {}


"""An object that auto updates itself. Use with the exchange and portfolio objects"""
import cloudpickle as cp
import json
import copy
import hashlib
import os
import time
import warnings
from copy import deepcopy
from redis import Redis
from hashlib import sha1
from crayons import magenta, yellow


from pprint import pprint

from funtime import Store
from loguru import logger

warnings.filterwarnings("ignore")



MONGO = os.getenv('MONGODB', "localhost")
PORTFOLIO_COLLECTION = os.getenv("PORTFOLIO_COLLECTION", "global")
REDIS = os.getenv('REDISHOST', "localhost")

# Sync using redis instead of price service. It's faster and requires less overhead.
redis_service = Redis(REDIS)


store = Store(MONGO).create_lib(PORTFOLIO_COLLECTION).get_store()

# Funtime declaration
db = store[PORTFOLIO_COLLECTION]


# Use the general object as a connection to funtime. Allows the user to store the objects necessary
# TODO: Convert into an event library and make it public
# TODO: Modify the format to better handle the change in store information. Should allow for a variation of locations
class AutoUpdateObject(object):
    def __init__(self, ID, TYPE, idhash=False):
        self.ID = ID
        if idhash == True:
            m = hashlib.md5()
            m.update(str(ID).encode('utf-8'))
            self.ID = m.hexdigest()
            
        
        self.composite_object = {
            'ID': self.ID,
            'type': TYPE,
            'id_hashes': []
        }
        
        self.query_item = {}


    def get_hash(self, query):
        return sha1(repr(sorted(query.items())).encode()).hexdigest()

    @logger.catch
    def safe_add(self, item):
        hashes = set(self.composite_object['id_hashes'])
        hashes.add(item)
        return list(hashes)
    
    @logger.catch
    def set_query_item(self, key, value):
        """Set an item you'd like to query to."""
        self.query_item[key] = value
    @logger.catch
    def clear_query_items(self):
        self.query_item = {}

    @logger.catch
    def set_id_hash(self, item):
        m = hashlib.md5()
        str_item = str(item).encode('utf-8')
        m.update(str_item)
        item_hash = m.hexdigest()
        self.composite_object['id_hashes'] = self.safe_add(item_hash)
    
    
    @logger.catch
    def merge(self, x):
        """ Merge Two Dictionaries together and update teh composite object together"""
        """ Used to update the entire dictionary at one time"""
        
        
        if isinstance(x, dict) == False:
            raise TypeError("Should be a dictionary")
        
        z = {**self.composite_object, **x}
        self.composite_object = z

    @logger.catch
    def set_item(self, key, value):
        self.composite_object[key] = value


    @logger.catch    
    def load(self):
        q = deepcopy(self.composite_object)
        q.pop('id_hashes', None)
        qq = {} 
        qq['ID'] = q['ID']
        qq['limit'] = 5
        qq['type'] = q['type']
        
        
        test_query = {
            "type": q['type'],
            "ID": q['ID'],
            **self.query_item
        }

        qhash = sha1(repr(sorted(test_query.items())).encode()).hexdigest()
        # qhash2 = self.get_hash(test_query)
        latest_red = redis_service.get(qhash)
        if latest_red is not None:
            self.composite_object = json.loads(latest_red)
            return

        item_list = list(db.query_sorted(test_query)) # Need to create a new kind of query in funtime.

        
        try:
            self.composite_object = item_list[0]
        except:
            pass
            # logger.error("There was no item available")
    


    @logger.catch    
    def load_wo(self):
        q = deepcopy(self.composite_object)
        q.pop('id_hashes', None)
        qq = {} 
        qq['limit'] = 5
        qq['type'] = q['type']
        
        
        qqq = {**qq, **self.query_item}
        test_query = {
            "type": q['type'],
            **self.query_item
        }
        qhash = sha1(repr(sorted(test_query.items())).encode()).hexdigest()
        
        latest_red = redis_service.get(qhash)
        if latest_red is not None:
            self.composite_object = json.loads(latest_red)
            return
        
        item_list = list(db.query_sorted(test_query)) # Need to create a new kind of query in funtime.
        # print(test)
        
        # item_list = list(db.query_latest(qqq)) # Need to create a new kind of query in funtime. 
        # logger.info(q['ID'])
        # print(item_list)
        try:
            self.composite_object = item_list[0]
        except:
            logger.error("There was no item available")

    # Timestamp should be ts if it's outside
    @logger.catch
    def update(self, timestamp=None):
        
        # Set to funtime
        if timestamp is None:
            self.composite_object['timestamp'] = time.time()
        else:
            self.composite_object['timestamp'] = timestamp
        i = self.composite_object
        
        q = deepcopy(self.composite_object)
        
        
        
        test_query = {
            "type": q['type'],
            "ID": q['ID'],
            **self.query_item
        }

        qhash = sha1(repr(sorted(test_query.items())).encode()).hexdigest()
        # redis_service.set(f"{qhash}:refresh", "placeholder", ex=10)
        # "
        
        savable = json.dumps(q)
        redis_service.set(qhash, savable, ex=10)
        

        # redis_service.set()
        db.store(i)
    
    @logger.catch
    def update_replace(self, timestamp=None):
        
        # Set to funtime
        rep_query = copy.copy(self.composite_object)
        if timestamp is None:
            self.composite_object['timestamp'] = time.time()
        else:
            self.composite_object['timestamp'] = timestamp
        i = self.composite_object
        
        remove_query = {
            "type": self.composite_object['type'],
            **self.query_item
        }
        
        q = deepcopy(self.composite_object)
        
        
        
        test_query = {
            "type": q['type'],
            "ID": q['ID'],
            **self.query_item
        }

        try:
            db.delete(remove_query)
            db.store(i)
            qhash = sha1(repr(sorted(test_query.items())).encode()).hexdigest()
            savable = json.dumps(q)
            redis_service.set(qhash, savable, ex=10)
        except:
            pass
    
    @logger.catch
    def get(self):
        return self.composite_object
    
    @logger.catch
    def lget(self):
        self.load()
        return self.composite_object
    
    @logger.catch
    def lget_all(self):
        q = deepcopy(self.composite_object)
        q.pop('id_hashes', None)
        test_query = {
            "type": q['type'],
            "ID": q['ID'],
            **self.query_item
        }
        qhash = sha1(repr(sorted(test_query.items())).encode()).hexdigest()
        # print(test_query)
        # print(magenta(qhash))
        # print(yellow(test_query.keys()))
        all_key = f"{qhash}:many"
        all_info = redis_service.get(all_key)
        if all_info is not None:
            loaded = json.loads(all_info)
            return loaded
        item_list = list(db.query_latest(test_query)) # Need to create a new kind of query in funtime.
        savable = json.dumps(item_list)
        redis_service.set(all_key, savable, ex=3)
        
        return item_list
    

    def create_key_base(self):
        pass


    @logger.catch
    def lget_all_update(self):
        # q = deepcopy(self.composite_object)
        # q.pop('id_hashes', None)
        test_query = {
            "type": self.composite_object['type'],
            "ID": self.composite_object['ID'],
            **self.query_item
        }
        # print(test_query)
        qhash = sha1(repr(sorted(test_query.items())).encode()).hexdigest()
        all_update_key = f"{qhash}:updatable"
        is_update = redis_service.get(all_update_key)

        if is_update is not None:
            # print(is_update)
            return
        redis_service.set(all_update_key, "True", ex=1.2)
        all_key = f"{qhash}:many"
        item_list = list(db.query_latest(test_query)) # Need to create a new kind of query in funtime.
        savable = json.dumps(item_list)
        redis_service.set(all_key, savable, ex=3)
        

        # return self.composite_object

    @logger.catch
    def lget_all_wo(self):
        q = deepcopy(self.composite_object)
        q.pop('id_hashes', None)
        test_query = {
            "type": q['type'],
            **self.query_item
        }

        qhash = sha1(repr(sorted(test_query.items())).encode()).hexdigest()
        all_key = f"{qhash}:many:wo"
        all_info = redis_service.get(all_key)
        if all_info is not None:
            loaded = json.loads(all_info)
            return loaded

        item_list = list(db.query_sorted(test_query)) # Need to create a new kind of query in funtime.
        savable = json.dumps(item_list)
        redis_service.set(all_key, savable, ex=3)
        
        return item_list


    @logger.catch
    def get_all(self):
        q = deepcopy(self.composite_object)
        q.pop('id_hashes', None)
        qq = {} 
        qq['ID'] = q['ID']
        qq['limit'] = 10000
        qq['type'] = q['type']
        qqq = {**qq, **self.query_item}
        
        qhash = sha1(repr(sorted(qqq.items())).encode()).hexdigest()
        all_key = f"{qhash}:many:get"
        all_info = redis_service.get(all_key)
        if all_info is not None:
            loaded = json.loads(all_info)
            return loaded

        item_list = list(db.query_latest(qqq))
        savable = json.dumps(item_list)
        redis_service.set(all_key, savable, ex=3)
        return item_list
    

    def get_all_wo(self):
        pass
    # Get without the id attached



from .data_handler import DataHandler

from .base_handlers.main import DataSet
from .base_handlers.runnable import RunnableSet
from .base_handlers.simulated import SimulatedData
from .base_handlers.backtest import BacktestDataset
from .base_handlers.price import PriceSet
from .base_handlers.live import LiveSet


import uuid
import pandas as pd
import maya
from jamboree import Jamboree, DBHandler

from loguru import logger
from typing import List, Dict, Any
from cytoolz.itertoolz import pluck
from crayons import yellow, blue, green, magenta
from link_jam_handle.data.pricing.simulated import SimulatedPriceHandler
from mlfinlab.portfolio_optimization.mean_variance import MeanVarianceOptimisation

"""
Note: Will remove soon. Going through a concept.
"""




class AllocationHandler(DBHandler):
    """Abstract handler that we use to keep track of information.
    """

    def __init__(self, limit=500, spacing=5, unit="hours"):
        super().__init__()
        self.entity = "allocation"
        self.required = {
            "user_id": str,
            "exchange": str,
            "episode": str,
            "live": bool
        }
        self._balance = 0
        self._limit = limit
        self._spacing = spacing
        self._unit = unit
        self.mean_variance_opt = MeanVarianceOptimisation()
    
    
        
    
    # -------------------------------------------------------------------
    # --------------------- Properties & Setters ------------------------
    # -------------------------------------------------------------------
    


    @property
    def limit(self):
        return self._limit
    
    @limit.setter
    def limit(self, limit):
        self._limit = limit
    

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, space):
        self._spacing = space



    @property
    def unit(self):
        return self._unit
    
    @unit.setter
    def unit(self, _unit):
        self._unit = _unit


    @property
    def assets(self) -> list:
        """ Get all of the assets that we're watching for the exchange """
        return self.latest_asset().get("assets", [])

    @property
    def allocation(self):
        return self.latest_allocation().get("allocation", {})
    
    @property
    def current_step(self):
        alt = {"detail": "steps"}
        count = self.count(alt=alt)
        return count
    
    @property
    def last_time(self):
        _late = self.latest_timestamp()
        
        return self.latest_timestamp().get("epoch", maya.now()._epoch)

    @property
    def should_rebalance(self) -> bool:
        """ Depending on the scenario, we determine if we should rebalance."""
        """
            if live:
                check if the determined amount of time has passed since we've rebalanced
            else:
                check if the predetermined number of steps have happened.
        """
        if self["live"]:
            current_epoch = maya.now()._epoch
            expected = maya.MayaDT(self.last_time).add(**{f"{self.unit}":self.spacing})._epoch
            # print(expected, current_epoch)
            if current_epoch >= expected:
                
                return True
        else:
            if self.current_step % self.spacing == 0:
                logger.info(green("Yes, we're going to rebalance", bold=True))
                return True
        return False

    def latest_asset(self):
        alt = {"detail": "assets"}
        asset = self.last(alt=alt)
        return asset
    
    def latest_allocation(self):
        alt = {"detail": "allocation"}
        asset = self.last(alt=alt)
        return asset
    

    def latest_timestamp(self):
        alt = {"detail": "timestamp"}
        asset = self.last(alt=alt)
        return asset


    def timestamp_count(self) -> int:
        """ Get the count of the asset adds we have. Not the number of assets. """
        alt = {"detail": "timestamp"}
        count = self.count(alt=alt)
        return count

    
    def asset_count(self) -> int:
        """ Get the count of the asset adds we have. Not the number of assets. """
        alt = {"detail": "assets"}
        count = self.count(alt=alt)
        return count
    


    def count_steps(self):
        alt = {"detail": "steps"}
        count = self.count(alt=alt)
        return count


    def add_assets(self, assets:List[str]):
        """ Add a list of assets to monitor. Used for tracking multiple assets inside of a portfolio. """
        if len(assets) == 0:
            return
        asset_list = self.assets
        for asset in assets:
            if asset not in asset_list:
                asset_list.append(asset)
        self.save_monitored_assets(asset_list)

    def add_asset(self, asset_name:str):
        """ Add an asset that will be monitored in an exchange"""
        if asset_name == "":
            return
        asset_list = self.assets
        asset_list.append(asset_name)
        self.save_monitored_assets(asset_list)
        
    
    def save_monitored_assets(self, assets:list):
        """ Save monitored assets """
        alt = {"detail": "assets"}
        monitored = {
            "assets": assets
        }
        self.save(monitored, alt=alt)
    
    def save_allocation(self, allocation:dict):
        """ Save monitored assets """
        alt = {"detail": "allocation"}
        monitored = {
            "allocation": allocation
        }
        self.save(monitored, alt=alt)
    
    def save_timestamp(self, _time=None):
        """ Save monitored assets """
        if _time is None:
            _time = maya.now()._epoch
        alt = {"detail": "timestamp"}
        monitored = {
            "epoch": _time
        }
        self.save(monitored, alt=alt)


    def save_step(self):
        alt = {"detail": "steps"}
        monitored = {}
        self.save(monitored, alt=alt)
    







    def _filter_assets(self, all_assets:dict) -> dict:
        assets = self.assets
        users_portfolio_to_rebalance = {}
        _all_assets = all_assets.get("assets", None)
        if _all_assets is None:
            return {}
        for asset in assets:
            current_asset = _all_assets.get(asset, None)
            if current_asset is not None:
                users_portfolio_to_rebalance[asset] = current_asset

        return users_portfolio_to_rebalance
    

    def _extract_frame(self, portfolio_assets:dict):
        time_is_filled = False
        time_index = []
        portfolio_close = {}
        for asset in portfolio_assets.keys():
            current_pricing = portfolio_assets.get(asset, [])
            portfolio_close[asset] = list(pluck("close", current_pricing))
            if time_is_filled == False:
                time_index = list(pluck("time", current_pricing))
                time_is_filled = True
        
        pandas_index = pd.to_datetime(time_index, unit='s')
        pclose = pd.DataFrame(portfolio_close, index=pandas_index)
        return pclose

    

    


    def _finish_step(self, is_rebalance=False):
        """ Finish update that we're not """
        if is_rebalance == True:
            # Give a timestamp of the last rebalance.
            print(yellow("saving timestamp"))
            self.save_timestamp()
        self.save_step()
        


    def step(self, all_assets:Dict[str, Any]) -> None:
        """ Accept all of the assets for a given allocation and begin"""
        

        if self.should_rebalance == True:
            users_portfolio_to_rebalance = self._filter_assets(all_assets)

            portfolio_keys = users_portfolio_to_rebalance.keys()
            if len(portfolio_keys) == 0:
                logger.info(magenta("There are no portfolio keys", bold=True))
                self._finish_step()
                return

            pclose = self._extract_frame(users_portfolio_to_rebalance)
            self.mean_variance_opt.allocate(pclose)
            current_allocation = self.mean_variance_opt.weights.to_dict('records')[0]
            logger.info(current_allocation)
            self.save_allocation(current_allocation)
            self._finish_step(is_rebalance=True)
            return
            # Set timestamp aka, the last time we did a rebalance.
        self._finish_step()


    def _reset_assets(self) -> None:
        if self.asset_count() == 0:
            logger.info(yellow("Setting the first assets"))
            asset_list = self.assets
            self.save_monitored_assets(asset_list)
        else:
            logger.info(blue("Already reset"))

    def _reset_timestamp(self):
        if self.timestamp_count() == 0:
            logger.info("We're setting a null timestamp. This way the rebalance will happen automatically upon the first step.")
            self.save_timestamp(_time=maya.when('2 years ago')._epoch)


    def reset(self):
        self._reset_assets()
        self._reset_timestamp()


def experiment_main():
    jam = Jamboree()
    allo_handler = AllocationHandler(unit="seconds", spacing=2)
    price_handler = SimulatedPriceHandler()

    allo_handler.event = jam
    price_handler.event = jam
    
    
    
    allo_handler["user_id"] = "1"
    allo_handler["episode"] = uuid.uuid4().hex
    allo_handler["exchange"] = f"binance-{uuid.uuid4().hex}"
    allo_handler["live"] = True
    
    asset_list = ["BTC", "ATL", "TRX", "ETH", "BCH", "XRP", "LTC", "EOS", "ADA", "XMR", "LINK", "HT"]
    # This live variable locks certain commands, and allows us to query faster. 
    price_handler['episode'] = uuid.uuid4().hex # set this to live if we're trying to use live prices.
    price_handler['exchange'] = "binance"
    price_handler['live'] = False
    price_handler['obs_num'] = 2000
    price_handler.add_assets(asset_list)
    price_handler.reset()


    for _ in range(10):
        allo_handler.reset()
    allo_handler.add_assets(["BTC", "ETH", "TRX", "BCH"])


    while True:

        obs = price_handler.step()

        # Get the observation number
        i = obs.get("step_num", 0)
        assets = obs.get("assets", {})
        done = obs.get("done", False)
        if done:
            logger.info(blue("Observation Complete"))
            break
        # print()
        # latest_price = price_handler.latest_price("BTC")
        # logger.info(yellow(latest_price['time']))
        # logger.info(red(price_handler.current_time))

        allo_handler.step(obs)
    

if __name__ == "__main__":
    experiment_main()


""" Get Aggregate Information """
from abc import ABC
from crayons import blue
from typing import Any, Dict
from link_jam_handle.data.base_handlers import RunnableSet
# Come back to work on this. It'll take a little while to finish this. It's classified as overbuilt unti lnecessary to do better. 

class BacktestDataset(RunnableSet, ABC):
    """ Backtest A Dataset"""
    def __init__(self, limit=500, seconds=0, minutes=0, hours=1, days=0, months=0, years=0) -> None:
        super().__init__(limit=limit, seconds=seconds, minutes=minutes, hours=hours, days=days, months=months, years=years)
        self._name = ""
        self._metadata = {}


    @property
    def name(self) -> str:
        return self.name

    @name.setter
    def name(self, name:str):
        self.name = name

    @property
    def metadata(self) -> dict:
        return self._metadata

    @metadata.setter
    def metadata(self, meta):
        self._metadata = meta

    def check(self):
        """ Check to see if a certain kind of dataset exists. """
        pass

    def load(self):
        """ Loads a dataset from the database and implants it into a backtest-able location if it's not already there. """
        print(blue(f"Load the dataset with the parameters ... {self._metadata}")) # would use the metadata to load from main dataset. 


    def latest(self) -> Dict[str, Any]:
        return {}

    def reset(self):
        """ Pull the data we need to pull """
        self.load()


import maya
import copy
from maya import MayaDT
from abc import ABC
from loguru import logger
from link_jam_handle.data.base_handlers import DataSet

class LiveSet(DataSet, ABC):
    """ Load and save live data. Should have a set of  """
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._parameters = {}
        self['opt_type'] = "live"
    

    @property
    def current_time(self):
        """ Get the timestamp of the price indicator. """
        return maya.now()._epoch

    
    

    def step(self):
        """ Step through simulated data """
        logger.error("Live data doesn't use the `step` paradigm")

    def reset(self):
        """ Run reset command to signify that we're loading information and"""
        logger.error("Live data doesn't use the `reset` paradigm")

    def close(self):
        logger.error("Live data doesn't use the `error` paradigm")


""" Abstract for all things related to simulations. Using OpenAI format to handle. """
from abc import ABC
from crayons import yellow
# from linkkt_hand.data.base import DataSet
from link_jam_handle.data.base_handlers import RunnableSet

class SimulatedData(RunnableSet, ABC):
    """ Simulated data has creation processes """
    def __init__(self, limit=500, **kwargs) -> None:
        super().__init__(limit=limit, **kwargs)
        self._parameters = {}
        self._is_generate = False

    @property
    def parameters(self):
        """ Parameters to we use to generate the data. """
        return self._parameters
    
    @parameters.setter
    def parameters(self, param:dict):
        """ Add items to the parameter dict. """
        keys = param.keys()
        for key in keys:
            self._parameters[key] = param[key]
    
    def _can_generate(self):
        """ Given the conditions we set, determine if we should generate data and sets a bool for the is_generate property to return. """
        print(yellow("Determining", bold=True))
    
    @property
    def is_generate(self) -> bool:
        """ Determine if we can generate"""
        self._can_generate()
        return self._is_generate

    def step(self):
        """ Step through runnable data """
        raise NotImplementedError

    def reset(self):
        """ Run reset command to signify that we're loading information and"""
        raise NotImplementedError

    def close(self):
        raise NotImplementedError




""" 
# BasePriceHandler

BasePriceHandler is an extension of DataHandler
"""



from jamboree import Jamboree
from loguru import logger
from crayons import yellow
from link_jam_handle.data import LiveSet
from link_jam_handle.data.pricing import PriceBaseSet
# TODO: Rethink design here and come back to this.

def is_acceptable_units(unit):
    units = ["seconds", "minutes", "hours", "days", "months", "years"]
    if unit not in units:
        return "minutes"
    return unit

class LivePriceHandler(LiveSet, PriceBaseSet):
    def __init__(self, limit=500, **kwargs):
        super().__init__(limit=limit, **kwargs)
        _units = kwargs.get("units", "minutes")
        _units = is_acceptable_units(_units)
        self.entity = "pricing"
        self['opt_type'] = "live"
        self['spacing'] = 1
        self['unit'] = _units
        self.asset_bars = {}

    
    def get_latest_many(self, symbol):
        pass
    """ 
        ----------------------------------------------------------------------------------
        ------------------------------- Counting Functions -------------------------------
        ----------------------------------------------------------------------------------
    """


    

    
    # """ ---------------------------------------------------------------------------
    #     ---------------------------- RL-Like Functions ---------------------------- 
    #     ---------------------------------------------------------------------------
    # """

    def step(self, _limit=2, alt={}):
        """
            # Step through the data set

            Step through the price dataset. The rules change based on the scenario.
        """
        current_step = self._step_count()
        if current_step == 0:
            # Use the first_limit here
            logger.info(yellow(f"Set the limit to the first limit: {self._first_limit}", bold=True))
            _limit = self._first_limit
        
        assets = self.assets
        
        swapped_assets = {
            "step_num": current_step,
            "assets": {},
            "done": self.is_next_observation
        }
        # Move the latest n assets over to the seen list

        for asset in assets:
            alt = {"asset": asset}
            self.swap_many(limit=_limit, alt=alt)
            swapped = self.query_many_swap(limit=self._first_limit, alt=alt)[::-1]
            swapped_assets["assets"][asset] = swapped
        
        swapped_assets["done"] = self.is_next_observation
        self.update_step()
        return swapped_assets
    
    def get_step(self):
        """
            # GET STEP
            ---
            Get all of the data for the latest step without pushing it forward.

            Only works when live is set to false. Should only use it when live is set to false. 
        """


    def reset(self):
        """ Reset should do nothing here. """
        logger.info("Resetting price information")

def main():
    asset_list = ["BTC", "ATL", "TRX", "ETH", "BCH", "XRP", "LTC", "EOS", "ADA", "XMR", "LINK", "HT"]
    price_handler = LivePriceHandler()
    jam = Jamboree()
    # This live variable locks certain commands, and allows us to query faster. 
    price_handler['episode'] = "live" # set this to live if we're trying to use live prices.
    price_handler['exchange'] = "binance"
    price_handler['live'] = True
    price_handler['obs_num'] = 1300
    print(price_handler['obs_num'])
    price_handler.event = jam
    price_handler.add_assets(asset_list)
    price_handler.reset()
    while True:
        obs = price_handler.step()

        # # Get the observation number
        # i = obs.get("step_num", 0)
        # assets = obs.get("assets", {})
        # done = obs.get("done", False)
        # if done:
        #     print(blue("Observation Complete"))
        #     break
        # print()
        # latest_price = price_handler.latest_price("BTC")
        # logger.info(yellow(latest_price['time']))
        # logger.info(red(price_handler.current_time))
        # logger.info(blue(assets['BTC'][-1]['time']))
    


if __name__ == "__main__":
    main()


import math
import numpy
import random
import decimal
import scipy.linalg
import numpy.random as nrand
import matplotlib.pyplot as plt
import pandas as pd
from datetime import datetime


class ModelParameters:
    """
    Encapsulates model parameters
    """

    def __init__(self,
                 all_s0, all_time, all_delta, all_sigma, gbm_mu,
                 jumps_lamda=0.0, jumps_sigma=0.0, jumps_mu=0.0,
                 cir_a=0.0, cir_mu=0.0, all_r0=0.0, cir_rho=0.0,
                 ou_a=0.0, ou_mu=0.0,
                 heston_a=0.0, heston_mu=0.0, heston_vol0=0.0):
        # This is the starting asset value
        self.all_s0 = all_s0
        # This is the amount of time to simulate for
        self.all_time = all_time
        # This is the delta, the rate of time e.g. 1/252 = daily, 1/12 = monthly
        self.all_delta = all_delta
        # This is the volatility of the stochastic processes
        self.all_sigma = all_sigma
        # This is the annual drift factor for geometric brownian motion
        self.gbm_mu = gbm_mu
        # This is the probability of a jump happening at each point in time
        self.lamda = jumps_lamda
        # This is the volatility of the jump size
        self.jumps_sigma = jumps_sigma
        # This is the average jump size
        self.jumps_mu = jumps_mu
        # This is the rate of mean reversion for Cox Ingersoll Ross
        self.cir_a = cir_a
        # This is the long run average interest rate for Cox Ingersoll Ross
        self.cir_mu = cir_mu
        # This is the starting interest rate value
        self.all_r0 = all_r0
        # This is the correlation between the wiener processes of the Heston model
        self.cir_rho = cir_rho
        # This is the rate of mean reversion for Ornstein Uhlenbeck
        self.ou_a = ou_a
        # This is the long run average interest rate for Ornstein Uhlenbeck
        self.ou_mu = ou_mu
        # This is the rate of mean reversion for volatility in the Heston model
        self.heston_a = heston_a
        # This is the long run average volatility for the Heston model
        self.heston_mu = heston_mu
        # This is the starting volatility value for the Heston model
        self.heston_vol0 = heston_vol0
    
def convert_to_returns(log_returns):
    """
    This method exponentiates a sequence of log returns to get daily returns.
    :param log_returns: the log returns to exponentiated
    :return: the exponentiated returns
    """
    return numpy.exp(log_returns)


def convert_to_prices(param, log_returns):
    """
    This method converts a sequence of log returns into normal returns (exponentiation) and then computes a price
    sequence given a starting price, param.all_s0.
    :param param: the model parameters object
    :param log_returns: the log returns to exponentiated
    :return:
    """
    returns = convert_to_returns(log_returns)
    # A sequence of prices starting with param.all_s0
    price_sequence = [param.all_s0]
    for i in range(1, len(returns)):
        # Add the price at t-1 * return at t
        price_sequence.append(price_sequence[i - 1] * returns[i - 1])
    return numpy.array(price_sequence)





def brownian_motion_log_returns(param):
    """
    This method returns a Wiener process. The Wiener process is also called Brownian motion. For more information
    about the Wiener process check out the Wikipedia page: http://en.wikipedia.org/wiki/Wiener_process
    :param param: the model parameters object
    :return: brownian motion log returns
    """
    sqrt_delta_sigma = math.sqrt(param.all_delta) * param.all_sigma
    return nrand.normal(loc=0, scale=sqrt_delta_sigma, size=param.all_time)


def brownian_motion_levels(param):
    """
    Returns a price sequence whose returns evolve according to a brownian motion
    :param param: model parameters object
    :return: returns a price sequence which follows a brownian motion
    """
    return convert_to_prices(param, brownian_motion_log_returns(param))


def geometric_brownian_motion_log_returns(param):
    """
    This method constructs a sequence of log returns which, when exponentiated, produce a random Geometric Brownian
    Motion (GBM). GBM is the stochastic process underlying the Black Scholes options pricing formula.
    :param param: model parameters object
    :return: returns the log returns of a geometric brownian motion process
    """
    assert isinstance(param, ModelParameters)
    wiener_process = numpy.array(brownian_motion_log_returns(param))
    sigma_pow_mu_delta = (param.gbm_mu - 0.5 * math.pow(param.all_sigma, 2.0)) * param.all_delta
    return wiener_process + sigma_pow_mu_delta


def geometric_brownian_motion_levels(param):
    """
    Returns a sequence of price levels for an asset which evolves according to a geometric brownian motion
    :param param: model parameters object
    :return: the price levels for the asset
    """
    return convert_to_prices(param, geometric_brownian_motion_log_returns(param))

def jump_diffusion_process(param):
    """
    This method produces a sequence of Jump Sizes which represent a jump diffusion process. These jumps are combined
    with a geometric brownian motion (log returns) to produce the Merton model.
    :param param: the model parameters object
    :return: jump sizes for each point in time (mostly zeroes if jumps are infrequent)
    """
    assert isinstance(param, ModelParameters)
    s_n = time = 0
    small_lamda = -(1.0 / param.lamda)
    jump_sizes = []
    for k in range(0, param.all_time):
        jump_sizes.append(0.0)
    while s_n < param.all_time:
        s_n += small_lamda * math.log(random.uniform(0, 1))
        for j in range(0, param.all_time):
            if time * param.all_delta <= s_n * param.all_delta <= (j + 1) * param.all_delta:
                # print("was true")
                jump_sizes[j] += random.normalvariate(param.jumps_mu, param.jumps_sigma)
                break
        time += 1
    return jump_sizes


def geometric_brownian_motion_jump_diffusion_log_returns(param):
    """
    This method constructs combines a geometric brownian motion process (log returns) with a jump diffusion process
    (log returns) to produce a sequence of gbm jump returns.
    :param param: model parameters object
    :return: returns a GBM process with jumps in it
    """
    assert isinstance(param, ModelParameters)
    jump_diffusion = jump_diffusion_process(param)
    geometric_brownian_motion = geometric_brownian_motion_log_returns(param)
    return numpy.add(jump_diffusion, geometric_brownian_motion)


def geometric_brownian_motion_jump_diffusion_levels(param):
    """
    This method converts a sequence of gbm jmp returns into a price sequence which evolves according to a geometric
    brownian motion but can contain jumps at any point in time.
    :param param: model parameters object
    :return: the price levels
    """
    return convert_to_prices(param, geometric_brownian_motion_jump_diffusion_log_returns(param))


def cox_ingersoll_ross_heston(param):
    """
    This method returns the rate levels of a mean-reverting cox ingersoll ross process. It is used to model interest
    rates as well as stochastic volatility in the Heston model. Because the returns between the underlying and the
    stochastic volatility should be correlated we pass a correlated Brownian motion process into the method from which
    the interest rate levels are constructed. The other correlated process is used in the Heston model
    :param param: the model parameters objects
    :return: the interest rate levels for the CIR process
    """
    # We don't multiply by sigma here because we do that in heston
    sqrt_delta_sigma = math.sqrt(param.all_delta) * param.all_sigma
    brownian_motion_volatility = nrand.normal(loc=0, scale=sqrt_delta_sigma, size=param.all_time)
    a, mu, zero = param.heston_a, param.heston_mu, param.heston_vol0
    volatilities = [zero]
    for i in range(1, param.all_time):
        drift = a * (mu - volatilities[i-1]) * param.all_delta
        randomness = math.sqrt(volatilities[i - 1]) * brownian_motion_volatility[i - 1]
        volatilities.append(volatilities[i - 1] + drift + randomness)
    return numpy.array(brownian_motion_volatility), numpy.array(volatilities)


def heston_model_levels(param):
    """
    NOTE - this method is dodgy! Need to debug!
    The Heston model is the geometric brownian motion model with stochastic volatility. This stochastic volatility is
    given by the cox ingersoll ross process. Step one on this method is to construct two correlated GBM processes. One
    is used for the underlying asset prices and the other is used for the stochastic volatility levels
    :param param: model parameters object
    :return: the prices for an underlying following a Heston process
    """
    assert isinstance(param, ModelParameters)
    # Get two correlated brownian motion sequences for the volatility parameter and the underlying asset
    # brownian_motion_market, brownian_motion_vol = get_correlated_paths_simple(param)
    brownian, cir_process = cox_ingersoll_ross_heston(param)
    brownian, brownian_motion_market = heston_construct_correlated_path(param, brownian)

    heston_market_price_levels = [param.all_s0]
    for i in range(1, param.all_time):
        drift = param.gbm_mu * heston_market_price_levels[i - 1] * param.all_delta
        vol = cir_process[i - 1] * heston_market_price_levels[i - 1] * brownian_motion_market[i - 1]
        heston_market_price_levels.append(heston_market_price_levels[i - 1] + drift + vol)
    return numpy.array(heston_market_price_levels), numpy.array(cir_process)


def heston_construct_correlated_path(param, brownian_motion_one):
    """
    This method is a simplified version of the Cholesky decomposition method for just two assets. It does not make use
    of matrix algebra and is therefore quite easy to implement.
    :param param: model parameters object
    :return: a correlated brownian motion path
    """
    # We do not multiply by sigma here, we do that in the Heston model
    sqrt_delta = math.sqrt(param.all_delta)
    # Construct a path correlated to the first path
    brownian_motion_two = []
    for i in range(param.all_time - 1):
        term_one = param.cir_rho * brownian_motion_one[i]
        term_two = math.sqrt(1 - math.pow(param.cir_rho, 2.0)) * random.normalvariate(0, sqrt_delta)
        brownian_motion_two.append(term_one + term_two)
    return numpy.array(brownian_motion_one), numpy.array(brownian_motion_two)


def get_correlated_geometric_brownian_motions(param, correlation_matrix, n):
    """
    This method can construct a basket of correlated asset paths using the Cholesky decomposition method
    :param param: model parameters object
    :param correlation_matrix: nxn correlation matrix
    :param n: the number of assets i.e. the number of paths to return
    :return: n correlated log return geometric brownian motion processes
    """
    assert isinstance(param, ModelParameters)
    decomposition = scipy.linalg.cholesky(correlation_matrix, lower=False)
    uncorrelated_paths = []
    sqrt_delta_sigma = math.sqrt(param.all_delta) * param.all_sigma
    # Construct uncorrelated paths to convert into correlated paths
    for i in range(param.all_time):
        uncorrelated_random_numbers = []
        for j in range(n):
            uncorrelated_random_numbers.append(random.normalvariate(0, sqrt_delta_sigma))
        uncorrelated_paths.append(numpy.array(uncorrelated_random_numbers))
    uncorrelated_matrix = numpy.matrix(uncorrelated_paths)
    correlated_matrix = uncorrelated_matrix * decomposition
    assert isinstance(correlated_matrix, numpy.matrix)
    # The rest of this method just extracts paths from the matrix
    extracted_paths = []
    for i in range(1, n + 1):
        extracted_paths.append([])
    for j in range(0, len(correlated_matrix)*n - n, n):
        for i in range(n):
            extracted_paths[i].append(correlated_matrix.item(j + i))
    return extracted_paths


def cox_ingersoll_ross_levels(param):
    """
    This method returns the rate levels of a mean-reverting cox ingersoll ross process. It is used to model interest
    rates as well as stochastic volatility in the Heston model. Because the returns between the underlying and the
    stochastic volatility should be correlated we pass a correlated Brownian motion process into the method from which
    the interest rate levels are constructed. The other correlated process is used in the Heston model
    :param param: the model parameters object
    :return: the interest rate levels for the CIR process
    """
    brownian_motion = brownian_motion_log_returns(param)
    # Setup the parameters for interest rates
    a, mu, zero = param.cir_a, param.cir_mu, param.all_r0
    # Assumes output is in levels
    levels = [zero]
    for i in range(1, param.all_time):
        drift = a * (mu - levels[i-1]) * param.all_delta
        # The main difference between this and the Ornstein Uhlenbeck model is that we multiply the 'random'
        # component by the square-root of the previous level i.e. the process has level dependent interest rates.
        randomness = math.sqrt(levels[i - 1]) * brownian_motion[i - 1]
        levels.append(levels[i - 1] + drift + randomness)
    return numpy.array(levels)


def ornstein_uhlenbeck_levels(param):
    """
    This method returns the rate levels of a mean-reverting ornstein uhlenbeck process.
    :param param: the model parameters object
    :return: the interest rate levels for the Ornstein Uhlenbeck process
    """
    ou_levels = [param.all_r0]
    brownian_motion_returns = brownian_motion_log_returns(param)
    for i in range(1, param.all_time):
        drift = param.ou_a * (param.ou_mu - ou_levels[i-1]) * param.all_delta
        randomness = brownian_motion_returns[i - 1]
        ou_levels.append(ou_levels[i - 1] + drift + randomness)
    return numpy.array(ou_levels)



def generate_super_price(base_price, _type, _time):
    delta = delta =1/(252*24)
    super_params = ModelParameters(
        all_s0=base_price,
        all_r0=0.001,
        all_time=_time,
        all_delta=delta,
        all_sigma=random.uniform(0.1, 0.8),
        gbm_mu=random.uniform(-0.3, 0.6),
        jumps_lamda=random.uniform(0.0071, 0.6),
        jumps_sigma=random.uniform(-0.03, 0.04),
        jumps_mu=random.uniform(-0.2, 0.2),
        cir_a=3.0,
        cir_mu=0.5,
        cir_rho=0.5,
        ou_a=3.0,
        ou_mu=0.5,
        heston_a=random.uniform(1, 5),
        heston_mu=random.uniform(0.156, 0.693),
        heston_vol0=0.06125
    )
    
    if _type.upper() == "GBM":
        m = geometric_brownian_motion_levels(super_params)
        return m
    elif _type.upper() == "HESTON":
        m = heston_model_levels(super_params)[0]
        return m
    elif _type.upper() == "MERTON":
        m = geometric_brownian_motion_jump_diffusion_levels(super_params)
        return m
    elif _type.upper() == "COX":
        m = cox_ingersoll_ross_levels(super_params)
        return m
    elif _type.upper() == "ORNSTEIN UHLENBECK":
        m = ornstein_uhlenbeck_levels(super_params)
        return m


""" 
# BasePriceHandler

BasePriceHandler is an extension of DataHandler
"""


import uuid
import dask
import random
import numpy as np
from jamboree import Jamboree
from loguru import logger
from crayons import (magenta, yellow, red, blue)

from link_jam_handle.utils import generate_super_price
from link_jam_handle.data import SimulatedData
from link_jam_handle.data.pricing import PriceBaseSet

class SimulatedPriceHandler(SimulatedData, PriceBaseSet):
    def __init__(self, limit=500, **kwargs):
        super().__init__(limit=limit, **kwargs)
        self.entity = "pricing"
        self['opt_type'] = "simulated"
        self.asset_bars = {}

    

    def latest_price(self, symbol:str):
        alt = {"asset": symbol}
        all_prices = self.last(alt=alt)
        return all_prices.get("close", 0.0)
    
    """ 
        ----------------------------------------------------------------------------------
        ------------------------------- Counting Functions -------------------------------
        ----------------------------------------------------------------------------------
    """


    def _generate_price_bars(self, name, _type, _len, base_price):
        price = generate_super_price(base_price, _type, _len)
        ps = len(price)
        vol_multiplier = random.uniform(2, 20)
        open_multiplier = np.absolute(np.random.normal(1, 0.05, size=ps))
        high_multiplier = np.absolute(np.random.normal(1.12, 0.05, size=ps))
        low_multiplier =  np.absolute(np.random.normal(0.92, 0.05, size=ps))
        noise = np.absolute(
            np.random.normal(1, 0.2, size=ps)
        )


        volume = (noise * price) * vol_multiplier
        _open = price * open_multiplier
        high = high_multiplier * price
        _low = low_multiplier * price

        bars = []
        current_time = self._start_time_maya
        
        time_list = list(range(ps))
        time_list_sorted = sorted(time_list, reverse=True)
        for _ in time_list_sorted:
            
            bar = {
                "open": float(_open[_]),
                "close": float(price[_]),
                "high": float(high[_]),
                "volume": float(volume[_]),
                "low": float(_low[_]),
                "time": float((current_time.add(
                                seconds=(self._seconds * _), minutes=(self._minutes * _), 
                                hours=(self._hours * _), days=(self._days * _), 
                                months=(self._months * _), years=(self._years * _))._epoch))
            }
            bars.append(bar)
        # # Create a bunch of bars here
        self.asset_bars[name] = bars

    


    def generate(self, starting_min=50, starting_max=2000, is_varied=True, excluded_assets=[], _len=4000):
        """ Generate price bars for assets """
        assets = self.assets
        if len(assets) == 0:
            logger.error(red("No assets found"))
            return
        
        for excluded in excluded_assets:
            assets.remove(excluded)
        
        # Generate bars
        mid_point = ((starting_min + starting_max)/2)
        std = (mid_point/4)

        
        dask_tasks = []
        for asset in assets:
            base_price = random.normalvariate(mid_point, std)
            # price_quote = f"Create prices for {asset}, {base_price}"
            _type = random.choice(["GBM", "HESTON"])
            dask_task = dask.delayed(self._generate_price_bars)(asset, _type, _len, base_price)
            dask_tasks.append(dask_task) 
        dask.compute(*dask_tasks)

        for asset in assets:
            logger.info("Saving Bar")
            self.save_bar_multi_bar(asset, self.asset_bars[asset])
        
    

    

    def _reset_generate_price(self):
        count = self.price_count()
        logger.info("Change the conditions here. The conditions should allow for dynamically adding assets")
        if count == 0:
            logger.info(magenta("Generating price points", bold=True))
            starting_min = self.data.get("starting_min", 50)
            starting_max = self.data.get("starting_max", 2000)
            is_varied = self.data.get("is_varied", True)
            excluded_assets = self.data.get("excluded_assets", [])
            obs_num = self.data.get("obs_num", 4000) # number of observations
            self.generate(starting_min=starting_min, starting_max=starting_max, is_varied=is_varied, excluded_assets=excluded_assets, _len=obs_num)
            self.save({"data": "blank"})


    @property
    def is_next_observation(self) -> bool:
        """ Check to see if there's a next observation for all coins. """
        assets = self.assets
        if len(assets) == 0: False
        confirmations = list(map(self._is_next_observation_single, assets))
        # logger.info(confirmations)
        return (not all(confirmations))


    def _is_next_observation_single(self, name:str):
        alt = {"asset": name}
        count = self.count(alt=alt)
        if count == 0: return False
        return True
    # """ ---------------------------------------------------------------------------
    #     ---------------------------- RL-Like Functions ---------------------------- 
    #     ---------------------------------------------------------------------------
    # """

    def step(self, _limit=1, alt={}):
        """
            # Step through the data set

            Step through the price dataset. The rules change based on the scenario.
        """
        current_step = self._step_count()
        if current_step == 0:
            # Use the first_limit here
            logger.info(yellow(f"Set the limit to the first limit: {self._first_limit}", bold=True))
            _limit = self._first_limit
        
        assets = self.assets
        
        swapped_assets = {
            "step_num": current_step,
            "assets": {},
            "done": self.is_next_observation
        }
        # Move the latest n assets over to the seen list

        for asset in assets:
            alt = {"asset": asset}
            self.swap_many(limit=_limit, alt=alt)
            swapped = self.query_many_swap(limit=self._first_limit, alt=alt)[::-1]
            swapped_assets["assets"][asset] = swapped
        
        swapped_assets["done"] = self.is_next_observation
        self.update_step()
        return swapped_assets
    
    def get_step(self):
        """
            # GET STEP
            ---
            Get all of the data for the latest step without pushing it forward.

            Only works when live is set to false. Should only use it when live is set to false. 
        """


    def reset(self):
        """ Resets the way of interacting with the price. """
        # super().reset()
        self._reset_generate_price()


def main():
    asset_list = ["BTC", "ATL", "TRX", "ETH", "BCH", "XRP", "LTC", "EOS", "ADA", "XMR", "LINK", "HT"]
    price_handler = SimulatedPriceHandler()
    jam = Jamboree()
    # This live variable locks certain commands, and allows us to query faster. 
    price_handler['episode'] = uuid.uuid4().hex # set this to live if we're trying to use live prices.
    price_handler['exchange'] = "binance"
    price_handler['live'] = False
    price_handler['obs_num'] = 1300
    print(price_handler['obs_num'])
    price_handler.event = jam
    price_handler.add_assets(asset_list)
    price_handler.reset()
    last_current_time = 0
    last_price_time = 0
    while True:

        obs = price_handler.step()

        # Get the observation number
        i = obs.get("step_num", 0)
        assets = obs.get("assets", {})
        done = obs.get("done", False)
        if done:
            logger.info(blue("Observation Complete"))
            break
        # print()
        latest_price = price_handler.latest_price("BTC")
        logger.info(yellow(latest_price['time']))
        logger.info(red(price_handler.current_time))
        logger.info(blue(assets['BTC'][-1]['time']))

        # print()

        # logger.info(latest_price['time'] - price_handler.current_time)
        # logger.info(assets['BTC'][-1]['time'] - price_handler.current_time)
        logger.info(price_handler.current_time - last_current_time)
        logger.info(latest_price['time'] - last_price_time)
        last_current_time = price_handler.current_time
        last_price_time = latest_price['time']


if __name__ == "__main__":
    main()


""" 
# BasePriceHandler

BasePriceHandler is an extension of DataHandler
"""


import copy
from typing import List, Dict, Any

# TODO: Rethink design here and come back to this.

from link_jam_handle.data import PriceSet


class PriceBaseSet(PriceSet):
    def __init__(self, limit=500, **kwargs):
        super().__init__(limit=limit, **kwargs)
        self.entity = "pricing"
        self.required = {
            "episode": str,
            "exchange": str,
            "live": bool
        }
        self.asset_bars = {}
    
    
    @property
    def assets(self) -> list:
        """ Get all of the assets that we're watching for the exchange """
        return self.latest_asset().get("assets", [])
    
    
    def latest_price(self, symbol):
        alt = {"asset": symbol}
        price_bar = self.last(alt=alt)
        return price_bar
    
    
    def latest_asset(self):
        alt = {"detail": "assets"}
        asset = self.last(alt=alt)
        return asset
    
    """ 
        ----------------------------------------------------------------------------------
        ------------------------------- Counting Functions -------------------------------
        ----------------------------------------------------------------------------------
    """
    
    
    def asset_count(self) -> int:
        """ Get the count of the asset adds we have. Not the number of assets. """
        alt = {"detail": "assets"}
        count = self.count(alt=alt)
        return count

    def price_count(self) -> int:
        count = self.count()
        return count
    
    
    """ 
        --------------------------------------------------------------------------------
        ------------------------------- Saving Functions -------------------------------
        --------------------------------------------------------------------------------
    """



    def add_assets(self, assets:List[str]):
        """ Add a list of assets to monitor. Used for tracking multiple assets inside of a portfolio. """
        if len(assets) == 0:
            return
        asset_list = self.assets
        for asset in assets:
            if asset not in asset_list:
                asset_list.append(asset)
        self.save_monitored_assets(asset_list)

    def add_asset(self, asset_name:str):
        """ Add an asset that will be monitored in an exchange"""
        if asset_name == "":
            return

        asset_list = self.assets
        asset_list.append(asset_name)
        asset_list = list(set(asset_list))
        self.save_monitored_assets(asset_list)
        
    
    def save_monitored_assets(self, assets:list):
        """ Save monitored assets """
        alt = {"detail": "assets"}
        monitored = {
            "assets": assets
        }
        self.save(monitored, alt=alt)
    
    
    
    def save_bar_multi_bar(self, name:str, bars:List[Dict]):
        """ Save multiple points of data """
        alt = {"asset": name}
        if len(bars) == 0:
            return
        
        query = copy.copy(self._query)
        episode = query.get("episode")
        exchange = query.get("exchange")
        print(f"Saving bars for {name}-{episode}-{exchange}")
        self.add_asset(name)
        self.save_many(bars, alt=alt) # This will yield a flag, because why not?

    def save_bar(self, name, bar:Dict[str, Any]) -> None:
        alt = {"asset": name}
        if len(bar.keys()) == 0:
            return
        
        self.add_asset(name)
        self.save(bar, alt=alt)
    
    def _reset_assets(self):
        """ Set the first asset """
        count = self.asset_count()
        if count == 0:
            # Save that we have no new assets
            self.save_monitored_assets([])
    
    
    
    """ ---------------------------------------------------------------------------
        ---------------------------- RL-Like Functions ---------------------------- 
        ---------------------------------------------------------------------------
    """

    def step(self):
        """
            # Step through the data set
            ---
            Step through the price dataset. The rules change based on the scenario.
        """
        
    
    def get_step(self):
        """
            # GET STEP
            ---
            Get all of the data for the latest step without pushing it forward.

            Only works when live is set to false. Should only use it when live is set to false. 
        """

    def reset(self):
        """ Resets the way of interacting with the price. """
        self._reset_assets()


""" Get Aggregate Information """
from abc import ABC
from typing import Any, Dict
from link_jam_handle.data.base_handlers import DataSet, PriceSet

# Come back to work on this. It'll take a little while to finish this. It's classified as overbuilt.

class AggregateDataset(ABC):
    def __init__(self) -> None:
        self._datasets = set()

    def raise_if_no_price(self):
        if not self._is_already_price():
            raise AttributeError("You didn't add any price information to read from")

    def _is_already_price(self) -> bool:
        """ Checks to see if there's already price information. We can only allow for one piece of price information. """
        if len(self._datasets) != 0:
            for _set in self._datasets:
                if isinstance(_set, PriceSet):
                    return True 
        return False

    @property
    def datasets(self):
        return self._datasets

    @datasets.setter
    def datasets(self, dataset:DataSet):
        """ Add a dataset to pull. """
        if not self._is_already_price():
            self._datasets.add(dataset)

    def latest(self) -> Dict[str, Any]:
        # Iterates through all datasets and calls their latest parameter
        return self.preprocess({})
    

    def latest_price(self, symbol:str):
        self.raise_if_no_price()
        for _set in self._datasets:
            if isinstance(_set, PriceSet):
                return _set.latest_price(symbol)
        return 0.0

    def _preprocess_rules(self):
        """ 
            A set of preprocessing rules that need to be called for any dataset. 
            Usually to match the records so they can be processed. 
        """
        raise NotImplementedError

    def preprocess(self, data:dict) -> Dict[str, Any]:
        """ Get all of the queried data and preprocess it for a single frame. """
        return {}
    

    def reset(self):
        """ Run a reset function """
        # Run through all datasets and reset them. LIve datasets will skip extra processing
        if len(self._datasets) == 0:
            raise IndexError("Datasets can't be empty")


import crayons as cy
from loguru import logger

# logger.add(sys.stdout, format="{time} - {level} - {message}")
def symbol_split(symbol):
    if not isinstance(symbol, str):
        return None
    
    s = symbol.split("/")
    if len(s) != 2:
        return None
    
    return s



def get_target_percentage(desired, rl_pct):
    # Get a percentage of the total amount to sell/buy
    assert 0.0 <= desired <= 1.0
    assert 0.0 <= rl_pct <= 1.0 
    return desired * rl_pct


def get_divergence_of_total(current_percentage, target_percentage):
    if target_percentage >= current_percentage:
        return 0.0
    return current_percentage - target_percentage


def get_percentage_of_sale(divergence):
    return 1-divergence


def get_sale_amount(asset_amount, percentage_to_sell):
    return asset_amount * percentage_to_sell

def get_sale_pct(rl_pct, desired, current_percentage):
    target_pct = get_target_percentage(rl_pct=rl_pct, desired=desired)
    diverge = get_divergence_of_total(current_percentage=current_percentage, target_percentage=target_pct)
    percent_sale = get_percentage_of_sale(diverge)
    return percent_sale


def buy_divergence(current_pct, target_pct):
    return target_pct-current_pct


def get_buy_pct(rl_pct, desired, current_percentage):
    # How much we're buying of the total portfolio
    target = get_target_percentage(desired=desired, rl_pct=rl_pct)
    buy_div = buy_divergence(current_pct=current_percentage, target_pct=target)
    return buy_div

def get_usd_pct(buy_pct, total_portfolio_value):
    return buy_pct*total_portfolio_value

class OrderInfo(object):
    def get_buy_pct(self, rl_percentage, desired, current_percentage):

        # How much we're buying of the total portfolio
        target = get_target_percentage(desired=desired, rl_pct=rl_percentage)
        buy_div = buy_divergence(current_pct=current_percentage, target_pct=target)
        return buy_div
    
    def get_sale_pct(self, rl_percentage, desired, current_percentage):
        target_pct = get_target_percentage(rl_pct=rl_percentage, desired=desired)
        diverge = get_divergence_of_total(current_percentage=current_percentage, target_percentage=target_pct)
        percent_sale = get_percentage_of_sale(diverge)
        
        # print(cy.magenta("Sales Percentage", bold=True))
        # print("---")
        # print(f"RL Percentage: {rl_percentage}")
        # print(f"RL Desired: {desired}")
        # print(f"Targeted Percentage: {target_pct}")
        # print(f"Current percentage of total: {current_percentage}")
        # print(f"The difference of percentage {diverge}" )
        # print(f"The percentage of token we're holding that we intend to sell: {percent_sale}")
        # print(rl_percentage, desired, target_pct, current_percentage, diverge, percent_sale)
        # print()
        return percent_sale
    
    def get_buy_usd(self, rl_percentage, desired, current_percentage, total_portfolio_value):
        # print(rl_percentage, desired, target_pct, current_percentage, diverge, percent_sale)
        buy_pct = get_buy_pct(rl_percentage, desired, current_percentage)
        buy_usd = get_usd_pct(buy_pct, total_portfolio_value)
        
        # print(cy.magenta("Sales Percentage", bold=True))
        # print("---")
        # print(f"RL Percentage: {rl_percentage}")
        # print(f"RL Desired: {desired}")
        # print(f"Current percentage of total: {current_percentage}")
        # print(f"Targeted Percentage: {buy_pct}")
        # print(f"The difference of percentage {buy_usd}" )
        return buy_usd


from uuid import uuid4
from jamboree import DBHandler, Jamboree

from abc import ABCMeta
from pebble.pool import ThreadPool
class DataHandler(DBHandler, metaclass=ABCMeta):
    def __init__(self, _limit=500) -> None:
        super().__init__()
        self.limit = _limit
        self.pool = ThreadPool()
    
    
    def get_latest(self, symbol:str, alt={}):
        raise NotImplementedError
    

    def _save_step(self):
        alt = {"details": "step"}
        data = {"count": "count"}
        self.save(data, alt=alt)


    def _step_count(self):
        alt = {"details": "step"}
        count = self.count(alt=alt)
        return count


    def _test_load(self, limit=10000):
        """ 
            We're going to test the swap idea here
            NOTE: Test Load 
        """
        

    def step(self, _limit=1, first_limit=500, alt={}):
        """ 
            STEP THROUGH DATA
            ---
            Get all the data related to what we'd deem as a step
            Parameters:
                - _limit: 
                    - The number of steps we're swapping over.
                - alt: 
                    - The specifics to the query
        """
        
        raise NotImplementedError


    def step_data(self):
        """ Get the most recent stepped through data"""
        raise NotImplementedError

    

if __name__ == "__main__":
    # We'll this swap works pretty damn well.
    jam = Jamboree()
    data_handle = DataHandler()
    data_handle.entity = "test_commands"
    data_handle.required = {"test": str, "episode": str}
    data_handle.event = jam
    data_handle['test'] = "test"
    data_handle['episode'] = uuid4().hex
    multiplier = 60


import maya
from abc import ABC
from jamboree import DBHandler
from pebble.pool import ThreadPool

class DataSet(DBHandler, ABC):
    """ Abstracted way of handling variations of datasets. """
    def __init__(self, **kwargs) -> None:
        super().__init__()
        self._name = ""
        self._metadata = {}
        self.pool = ThreadPool()
        self._start_time_maya = maya.now()
        self._current_time = None

        self._seconds = kwargs.get("seconds", 0)
        self._minutes = kwargs.get("minutes", 0)
        self._hours = kwargs.get("hours", 1)
        self._days = kwargs.get("days", 0)
        self._months = kwargs.get("months", 0)
        self._years = kwargs.get("years", 0)
        self._first_limit = kwargs.get("first_limit", 500)

    @property
    def current_time(self):
        """ Get the timestamp of the price indicator. """
        raise NotImplementedError
        

    def get_latest(self, symbol:str, alt={}):
        raise NotImplementedError

    @property
    def name(self) -> str:
        return self.name

    @name.setter
    def name(self, name:str):
        self.name = name

    def latest(self, alt:dict={}):
        raise NotImplementedError



""" # Simulation Base Class
    ---
    Use simulation base class for all things simulation. Uses step architecture.
"""

from .main import DataSet
from .runnable import RunnableSet
from .simulated import SimulatedData
from .backtest import BacktestDataset
from .price import PriceSet